import base64
import datetime
import json
from math import floor

from gcpjwt.jwt_signer import JWTSigner


class JWT:
    def __init__(self, signer: JWTSigner, ring: str, key: str, hash: str = 'sha256'):
        self._signer = signer
        self._ring = ring
        self._key = key
        self._hash = hash

        self._iss = None
        self._exp = None
        self._sub = None
        self._aud = None
        self._iat = None
        self._nbf = None
        self._jti = None

        self._custom_claims = {}

    def utc_now(self):
        epoch = datetime.datetime(1970, 1, 1)
        return floor((datetime.datetime.utcnow() - epoch).total_seconds())

    @property
    def iss(self):
        return self._iss

    @iss.setter
    def iss(self, value: str):
        self._iss = value

    @property
    def exp(self):
        return self._exp

    @exp.setter
    def exp(self, value: int):
        self._exp = value

    @property
    def sub(self):
        return self._sub

    @sub.setter
    def sub(self, value: str):
        self._sub = value

    @property
    def aud(self):
        return self._aud

    @aud.setter
    def aud(self, value: str):
        self._aud = value

    @property
    def iat(self):
        return self._iat

    @iat.setter
    def iat(self, value: str):
        self._iat = value

    @property
    def nbf(self):
        return self._nbf

    @nbf.setter
    def nbf(self, value: int):
        self._nbf = value

    @property
    def jti(self):
        return self._jti

    @jti.setter
    def jti(self, value):
        self._jti = value

    def get_claim(self, name: str):
        return self._custom_claims[name]

    def set_claim(self, name: str, value):
        self._custom_claims[name] = value

    @property
    def payload(self):
        payload = {}
        if self.iss is not None:
            payload['iss'] = self.iss

        if self.iat is not None:
            payload['iat'] = self.iat
        else:
            payload['iat'] = self.utc_now()

        if self.exp is not None:
            payload['exp'] = self.exp
        else:
            payload['exp'] = payload['iat'] + (60 * 60 * 2)

        if self.sub is not None:
            payload['sub'] = self.sub
        if self.aud is not None:
            payload['aud'] = self.aud
        if self.nbf is not None:
            payload['nbf'] = self.nbf
        if self.jti is not None:
            payload['jti'] = self.jti

        return {**payload, **self._custom_claims}

    @property
    def signature(self):
        payload_bytes = json.dumps(self.payload).encode('utf-8')
        return self._signer.sign(self._ring, self._key, payload_bytes, self._hash)

    def token(self):
        payload = json.dumps(self.payload).encode('utf-8')

        sign_result = self.signature
        alg = sign_result[0]
        signature = sign_result[1]

        header = json.dumps({
            'alg': alg + self._hash,
            'typ': 'JWT'
        }).encode('utf-8')

        return '{}.{}.{}'.format(
            base64.b64encode(header).decode('utf-8'),
            base64.b64encode(payload).decode('utf-8'),
            base64.b64encode(signature).decode('utf-8')
        )

    def verify(self, token: str, force_latest: bool = False):
        broken = token.split('.')
        payload = base64.b64decode(broken[1].encode('utf-8'))
        signature = base64.b64decode(broken[2].encode('utf-8'))

        dict_payload = json.loads(payload)
        if 'exp' in dict_payload and dict_payload['exp'] <= self.utc_now():
            return 1

        valid = self._signer.verify(self._ring, self._key, payload, signature, force_latest, self._hash)[1]
        if not valid:
            return 2

        return 0



import hashlib

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import ec, rsa, utils
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey

from google.cloud.kms_v1 import KeyManagementServiceClient
from google.cloud.kms_v1.gapic import enums as kms_enums


class JWTSigner:
    def __init__(self, client: KeyManagementServiceClient, project: str, location: str):
        self.client = client
        self.project = project
        self.location = location

        self._hash_map = {
            'sha256': lambda x: hashlib.sha256(x).digest(),
            'sha384': lambda x: hashlib.sha384(x).digest(),
            'sha512': lambda x: hashlib.sha512(x).digest()
        }

        self._crypto_hash_map = {
            'sha256': hashes.SHA256(),
            'sha384': hashes.SHA384(),
            'sha512': hashes.SHA512()
        }

    def _get_key(self, key):
        pub_data = self.client.get_public_key(key.name).pem.encode('utf-8')
        return serialization.load_pem_public_key(pub_data, default_backend())

    def _verify(self, key, hash: str, signature: bytes, digest: bytes):
        if isinstance(key, EllipticCurvePublicKey):
            key.verify(signature, digest,
                       signature_algorithm=ec.ECDSA(utils.Prehashed(self._crypto_hash_map[hash])))
        else:
            key.verify(signature, digest, algorithm=self._crypto_hash_map[hash])

    def sign(self, ring: str, key: str, data: bytes, hash: str = 'sha256'):
        key_path = self.client.crypto_key_path(self.project, self.location, ring, key)
        enabled_keys = [e for e in list(self.client.list_crypto_key_versions(key_path))
                        if e.state == kms_enums.CryptoKeyVersion.CryptoKeyVersionState.ENABLED]
        latest = sorted(enabled_keys, key=lambda e: e.create_time.seconds, reverse=True)[0]
        type = kms_enums.CryptoKeyVersion.CryptoKeyVersionAlgorithm(latest.algorithm).name.split('_')[0]
        return type, self.client.asymmetric_sign(latest.name, {hash: self._hash_map[hash](data)}).signature

    def verify(self, ring: str, key: str, payload: bytes, signature: bytes, force_latest: bool = False, hash: str = 'sha256'):
        key_path = self.client.crypto_key_path(self.project, self.location, ring, key)
        enabled_keys = [e for e in list(self.client.list_crypto_key_versions(key_path))
                        if e.state == kms_enums.CryptoKeyVersion.CryptoKeyVersionState.ENABLED]
        if len(enabled_keys) == 0:
            return False

        if force_latest:
            latest = sorted(enabled_keys, key=lambda e: e.create_time.seconds, reverse=True)[0]
            public_key = self._get_key(latest)
            digest = self._hash_map[hash](payload)
            type = kms_enums.CryptoKeyVersion.CryptoKeyVersionAlgorithm(latest.algorithm).name.split('_')[0]

            try:
                if isinstance(key, EllipticCurvePublicKey):
                    self._verify(public_key, hash, signature, digest)
                return type, True
            except InvalidSignature:
                return type, False
        else:
            type = None
            for key in enabled_keys:
                public_key = self._get_key(key)
                digest = self._hash_map[hash](payload)
                type = kms_enums.CryptoKeyVersion.CryptoKeyVersionAlgorithm(key.algorithm).name.split('_')[0]

                try:
                    self._verify(public_key, hash, signature, digest)
                    return type, True
                except InvalidSignature:
                    continue
            return type, False



import json

from typing import List, Tuple
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()


def simple_setup(info, data_files: List[Tuple[str, List[str]]] = [], python_version: str = '>=3.6', excludes: list = []):
    with open('resources/requirements.txt') as fr:
        requirements = [x for x in fr.readlines() if not x.startswith('--extra-index-url')]
        links = [x[18:] for x in fr.readlines() if x.startswith('--extra-index-url')]

    data_files.append(('resources', ['resources/setup.json', 'resources/requirements.txt']))
    setup(
        **info,

        long_description=long_description,
        long_description_content_type="text/markdown",

        install_requires=requirements,
        python_requires=python_version,
        dependency_links=links,

        packages=find_packages(exclude=excludes),
        data_files=data_files,
        include_package_data=True
    )


simple_setup(json.load(open('resources/setup.json')))
